﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CatanGame.Model
{
    public class CatanBoard
    {
        #region Private fields
        private readonly static Dictionary<ResourceType, int> resourcePossibilities = new Dictionary<ResourceType, int> {
            { ResourceType.None, 1 },
            { ResourceType.Grain, 4 },
            { ResourceType.Lumber, 4 },
            { ResourceType.Wool, 4 },
            { ResourceType.Brick, 3 },
            { ResourceType.Ore, 3 },
        };

        private readonly static Dictionary<int, int> numPossibilities = new Dictionary<int, int>
        {
            { 2, 1 },
            { 3, 2 },
            { 4, 2 },
            { 5, 2 },
            { 6, 2 },
            { 8, 2 },
            { 9, 2 },
            { 10, 2 },
            { 11, 2 },
            { 12, 1 },
        };

        private Player[] _players;
        private NodeMatrix _nodes;
        private RoadMatrix _roads;
        private Field[,] _fields;
        private int _currentPlayer;
        private static Random rand = new Random();
        private bool _initialTurn;
        private int _turnIncrement;
        private bool _rolled;
        private bool _robberPositionChanged;
        private int _maxLengthRoadOwnerId;
        private int _maxRoadLength;
        private int _maxKnightCardOwnerId;
        private int _maxKnightCardCount;
        #endregion

        #region Public properties
        /// <summary>
        /// Gets the players playing the game.
        /// </summary>
        public Player[] Players { get => _players; }
        /// <summary>
        /// Gets the nodes of the board.
        /// </summary>
        public NodeMatrix Nodes { get => _nodes; }
        /// <summary>
        /// Gets the roads of the board.
        /// </summary>
        public RoadMatrix Roads { get => _roads; }
        /// <summary>
        /// Gets the fields of the board.
        /// </summary>
        public Field[,] Fields { get => _fields; }
        /// <summary>
        /// Gets or sets the position of the robber.
        /// </summary>
        public (int, int) RobberPosition { get; set; }

        /// <summary>
        /// Gets the longest read length.
        /// </summary>
        public int MaxRoadLength { get => _maxRoadLength; }
        /// <summary>
        /// Gets the owner of the longest road.
        /// </summary>
        public int LongestRoadOwnerId { get => _maxLengthRoadOwnerId; }
        /// <summary>
        /// Gets the owner of the max knight card.
        /// </summary>
        public int MaxKnightCardOwnerId { get => _maxKnightCardOwnerId; }

        /// <summary>
        /// Gets or sets whether the dice have been rolled in the current turn.
        /// </summary>
        public bool Rolled { get => _rolled; set => _rolled = value; }

        /// <summary>
        /// Gets or sets the size of the board.
        /// </summary>
        public int Size { get; private set; }

        /// <summary>
        /// Gets whether we are still in the initial phase of the game.
        /// </summary>
        public bool InitialTurn { get => _initialTurn; }

        /// <summary>
        /// Gets whether the robber was placed or not.
        /// </summary>
        public bool RobberPlaced { get => _robberPositionChanged; }

        /// <summary>
        /// Gets the id of the current player.
        /// </summary>
        public int CurrentPlayerId { get => _currentPlayer; }

        /// <summary>
        /// Gets or sets the random generator.
        /// </summary>
        public static Random Rand { get => rand; set => rand = value; }

        /// <summary>
        /// Gets the hexagon count of the board.
        /// </summary>
        public int HexagonCount { get => Size * Size - (Size / 2 + 1) * Size / 2; }

        /// <summary>
        /// Gets the vertex count of the board.
        /// </summary>
        public int VertexCount { get => HexagonCount * 2 + Size * 3 + 1; }

        /// <summary>
        /// Gets the current player.
        /// </summary>
        public Player CurrentPlayer { get => _players[CurrentPlayerId]; }

        #endregion

        #region Public constructors
        public CatanBoard(int size = 7)
        {
            if (size % 2 == 0)
                throw new InvalidOperationException("size can't be an even number!");
            Size = size;

            _maxLengthRoadOwnerId = -1;
            _maxRoadLength = 0;
            _maxKnightCardOwnerId = -1;
            _maxKnightCardCount = 0;

            InitializeNodes();
            InitializeFields();
            InitializeRoads();
            InitializePlayers();

        }
        #endregion

        #region Public methods

        /// <summary>
        /// Buys a card for the given player.
        /// </summary>
        /// <param name="p">The player.</param>
        /// <returns>The card type that has been bought.</returns>
        public CardType BuyCard(Player p)
        {
            if (p.CanBuyDevelopmentCard && _rolled && _robberPositionChanged && !_initialTurn)
            {
                int cardType = rand.Next(0, 2);
                p.DevelopmentCards[(CardType)cardType] += 1;
                if ((CardType)cardType == CardType.VictoryPoint)
                    p.Score++;
                MaxKnightCard();
                p.DevelopmentCardBought();
                return (CardType)cardType;
            }
            else
            {
                return CardType.None;
            }
        }

        /// <summary>
        /// Places a road for the player between the nodes with the given coordinates.
        /// </summary>
        /// <param name="coords">The coordinates of the nodes.</param>
        /// <param name="p">The player.</param>
        /// <returns>True if the road was succesfully built, false otherwise.</returns>
        public bool PlaceRoad(RoadCoordinates coords, Player p)
        {
            if (!CanPlaceRoad(coords, p))
                return false;

            _roads.Add(coords, p.Id);
            p.RoadBuilt(_initialTurn);

            if (_initialTurn)
                StepGame();

            MaxLengthRoadSearch();

            return true;
        }

        /// <summary>
        ///     Calculates available roads for given player and returns them.
        /// </summary>
        /// <param name="p">The player.</param>
        /// <returns>Available roads for the given player.</returns>
        public List<RoadCoordinates> EnabledRoads(Player p)
        {
            List<RoadCoordinates> answer = new();
            if (_rolled && _robberPositionChanged)
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    for (int d = 0; d < 2; d++)
                        foreach (RoadCoordinates road in EnabledRoadsFromNode(p, new NodeCoordinates(r, q, (Direction)d)))
                            answer.Add(road);
                    
            return answer;
        }

        /// <summary>
        /// Calculates available nodes for given player and returns them.
        /// </summary>
        /// <param name="p">The player.</param>
        /// <returns>Available nodes for the given player.</returns>
        public List<NodeCoordinates> EnableNodes(Player p)
        {
            List<NodeCoordinates> answer = new();
            if (_rolled && _robberPositionChanged)
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    for (int d = 0; d < 2; d++)
                        if (CanPlaceSettlement(new NodeCoordinates(r, q, (Direction)d), p))
                            answer.Add(new NodeCoordinates(r, q, (Direction)d));
            return answer;
        }

        /// <summary>
        /// Calculates the available roads for the given player from the given node.
        /// </summary>
        /// <param name="p">The player.</param>
        /// <param name="current">The coordinates of the node.</param>
        /// <returns>The available roads for the given player from the given node.</returns>
        public List<RoadCoordinates> EnabledRoadsFromNode(Player p, NodeCoordinates current)
        {
            List<RoadCoordinates> answer = new List<RoadCoordinates>();
            List<NodeCoordinates> neighbors = GetNeighborCoords(current);
            foreach (NodeCoordinates neighbor in neighbors)
            {
                    if (Nodes.NodeExists(current) && Nodes.NodeExists(neighbor))
                    {
                        RoadCoordinates roadToPlace = new RoadCoordinates(current, neighbor);
                        if (!_roads.IsPlacedRoad(roadToPlace) && CanPlaceRoad(roadToPlace, p))
                            answer.Add(roadToPlace);
                    }
            }
            return answer;
        }

        /// <summary>
        /// Places a settlement to the given node.
        /// </summary>
        /// <param name="node">The coordinates of the node.</param>
        /// <param name="p">The player.</param>
        /// <returns>True if the placement was succesful, false otherwise.</returns>
        public bool PlaceSettlement(NodeCoordinates node, Player p)
        {
            if (CanPlaceSettlement(node, p))
            {
                House building = Nodes[node].Building;
                building.Type = HouseType.Settlement;
                building.Owner = p;
                p.AddedSettlement(_initialTurn);
                if (!_initialTurn)
                    Players[CurrentPlayerId].Score++;
                return true;
            }
            return false;
        }

        /// <summary>
        /// Upgrades a node to a city
        /// </summary>
        /// <param name="node">The coordinates of the node.</param>
        /// <param name="p">The player.</param>
        /// <returns>True if the upgrade was succesful, false otherwise.</returns>
        public bool UpgradeToCity(NodeCoordinates node, Player p)
        {
            if (CanUpgradeToCity(node, p))
            {
                Nodes[node].Building.Type = HouseType.City;
                p.UpgradedToCity();
                Players[CurrentPlayerId].Score++;
                return true;
            }
            return false;
        }

        /// <summary>
        /// Steps the game to the next state.
        /// </summary>
        public void StepGame()
        {
            if (!_robberPositionChanged)
                return;
            
            if (_rolled || _initialTurn)
                _currentPlayer += _turnIncrement;
            if (!_initialTurn)
                _rolled = false;
            if (_currentPlayer == _players.Length)
            {
                if (_initialTurn)
                {
                    _turnIncrement *= -1;
                    _currentPlayer += _turnIncrement;
                }
                else
                    _currentPlayer = 0;
            }
            if (_currentPlayer == -1)
            {
                _initialTurn = false;
                SendResources();
                _rolled = false;
                _turnIncrement *= -1;
                _currentPlayer += _turnIncrement;
            }
        }

        /// <summary>
        /// Performs a trade.
        /// </summary>
        /// <param name="offers">The resources offered by the player.</param>
        /// <param name="requests">The resources requested by the player.</param>
        /// <returns>True if the player could make the trade, false otherwise.</returns>
        public bool Trade(List<(ResourceType OfferedResource, int Quantity)> offers, List<(ResourceType RequestedResource, int Quantity)> requests)
        {
            if (CanTrade(offers, requests))
            {
                foreach (var offer in offers)
                    CurrentPlayer.Resources[offer.OfferedResource] -= offer.Quantity;
                foreach (var request in requests)
                    CurrentPlayer.Resources[request.RequestedResource] += request.Quantity;
                return true;
            }
            else
                return false;
        }

        /// <summary>
        /// Checks if the trade can be made.
        /// </summary>
        /// <param name="offers">The resources offered by the player.</param>
        /// <param name="requests">The resources requested by the player.</param>
        /// <returns>True if a trade can be made, false otherwise.</returns>
        public bool CanTrade(List<(ResourceType OfferedResource, int Quantity)> offers, List<(ResourceType RequestedResource, int Quantity)> requests)
        {
            return IsValidOffer(offers)
                && offers.Select(offer => offer.Quantity).Sum() == requests.Select(request => request.Quantity).Sum() * 3;
        }

        /// <summary>
        /// Checks if the current player can request any more resources.
        /// </summary>
        /// <param name="offers">The resources offered by the player.</param>
        /// <param name="requests">The resources requested by the player.</param>
        /// <returns>True if the player can request more resources, false otherwise.</returns>
        public bool CanRequest(List<(ResourceType OfferedResource, int Quantity)> offers, List<(ResourceType RequestedResource, int Quantity)> requests)
        {
            return IsValidOffer(offers)
                && offers.Select(offer => offer.Quantity).Sum() > requests.Select(request => request.Quantity).Sum() * 3;
        }

        /// <summary>
        /// Moves the robber.
        /// </summary>
        /// <param name="coords">The position where the robber should be moved to.</param>
        /// <returns>True if the robber was moved, false on failure.</returns>
        public bool MoveRobber((int r, int q) coords)
        {
            if (RobberPosition == coords || _robberPositionChanged || _initialTurn)
                return false;
            RobberPosition = (coords.r, coords.q);
            _robberPositionChanged = true;
            return true;
        }

        /// <summary>
        /// Calculates the possible positions for the robber.
        /// </summary>
        /// <returns>The possible positions for the robber.</returns>
        public List<(int R, int Q)> PossibleRobberPositions()
        {
            List<(int R, int Q)> result = new();
            for (int r = 0; r < Size; ++r)
            {
                for (int q = 0; q < Size; ++q)
                {
                    if (Fields[r, q] is not null && Fields[r, q].ResourceType != ResourceType.Sea && (r, q) != RobberPosition)
                        result.Add((r, q));
                }
            }
            return result;
        }

        /// <summary>
        /// Rolls the dice.
        /// </summary>
        /// <returns>The result of the roll.</returns>
        public (int, int) Roll()
        {
            
            (int FirstDice, int SecondDice) result = (rand.Next(1, 7), rand.Next(1, 7));

            if (result.FirstDice + result.SecondDice == 7)
                _robberPositionChanged = false;

            _rolled = true;
            return result;
        }

        /// <summary>
        /// Sends the resources of the fields with the given number to the players who have settlements next to them.
        /// </summary>
        /// <param name="num">The field number that got rolled.</param>
        public void SendResources(int num)
        {
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    if (_fields[r, q] != null &&_fields[r,q].FieldNum == num && (r, q) != RobberPosition)
                        _fields[r, q].SendResources();
        }

        /// <summary>
        /// Sends the resources of all the fields to the players who have settlements next to them.
        /// This is useful to call after the end of the initial round.
        /// </summary>
        public void SendResources()
        {
            for (int r = 0; r < Size; r++)
                for (int q = 0; q < Size; q++)
                    if (_fields[r, q] != null && (r, q) != RobberPosition)
                        _fields[r, q].SendResources();
        }

        /// <summary>
        /// Checks if a road can be placed between the nodes with the given coordinates.
        /// </summary>
        /// <param name="coords">The coordinates of the nodes.</param>
        /// <param name="p">The player.</param>
        /// <returns>True if the road can be placed, false otherwise.</returns>
        public bool CanPlaceRoad(RoadCoordinates coords, Player p)
        {
            bool answer = false;

            if (Roads.IsPlacedRoad(coords))
                return false;

            if (!GetNeighborCoords(coords.From).Contains(coords.To))
                return false;

            if (_initialTurn)
            {
                    answer |= Nodes.NodeExists(coords.From)
                        && Nodes[coords.From].Building.HasOwner
                        && Nodes[coords.From].Building.Owner.Id == p.Id;
                    
                    answer |= Nodes.NodeExists(coords.To)
                        && Nodes[coords.To].Building.HasOwner
                        && Nodes[coords.To].Building.Owner.Id == p.Id;
            }

            Predicate<Road> isConnected = (neighboringRoad) =>
                    neighboringRoad.PlayerId == p.Id
                    && (Nodes[neighboringRoad.Coordinates.From].Building.Type == HouseType.NoBuilding
                        || Nodes[neighboringRoad.Coordinates.From].Building.Owner.Id == p.Id);
            
            answer |= Roads.RoadsFromNode(coords.From).Exists(isConnected)
                   || Roads.RoadsFromNode(coords.To).Exists(isConnected);

            answer &= p.CanBuildRoad || _initialTurn;

            return answer;
        }


        /// <summary>
        /// Checks if a settlement can be placed at the node with given coordinates.
        /// </summary>
        /// <param name="node">The coordinates of the node.</param>
        /// <param name="p">The player.</param>
        /// <returns>True if a settlement can be placed, false otherwise.</returns>
        public bool CanPlaceSettlement(NodeCoordinates node, Player p)
        {
            if (_currentPlayer != p.Id)
                return false;

            List<NodeCoordinates> neighbors = GetNeighborCoords(node);
            bool answer = Nodes.NodeExists(node) 
                && Nodes[node].Building.Type == HouseType.NoBuilding 
                && (p.CanBuildSettlement || _initialTurn);

            answer &= neighbors.All((neighbor) =>
                !Nodes.NodeExists(neighbor) 
                || Nodes[neighbor].Building.Type == HouseType.NoBuilding);

            bool hasNeighboringRoad = Roads.RoadsFromNode(node).Any((road) => road.PlayerId == p.Id);

            answer &= hasNeighboringRoad || _initialTurn;

            return answer;
        }

        /// <summary>
        /// Checks if a player can upgrade the node with given coordinates to a city.
        /// </summary>
        /// <param name="node">The coordinates of the node.</param>
        /// <param name="p">The player.</param>
        /// <returns>True if the node can be upgraded to a city, false otherwise.</returns>
        public bool CanUpgradeToCity(NodeCoordinates node, Player p)
        {
            return !_initialTurn && Nodes[node].Building.Owner == p && p.CanBuildCity;
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Initializes the fields in the game.
        /// </summary>
        private void InitializeFields()
        {
            _fields = new Field[Size, Size];
            Dictionary<ResourceType, int> possibleResources = new(resourcePossibilities);
            Dictionary<int, int> possibleNums = new(numPossibilities);

            int innerBoardSize = Size - 2;
            for(int q = 0; q < Size; ++q)
            {
                for (int r = 0; r < Size; ++r)
                {
                    if (Math.Abs(q + r - Size + 1) < 1 + Size / 2) 
                    {
                        if (IsInnerBoard(r, q))
                        {
                            _fields[r, q] = new Field(new List<Node> {
                                Nodes[r,     q,      Direction.Bottom],
                                Nodes[r,     q,      Direction.Top],
                                Nodes[r - 1, q + 1,  Direction.Bottom],
                                Nodes[r - 1, q,      Direction.Bottom],
                                Nodes[r + 1, q - 1,  Direction.Top],
                                Nodes[r + 1, q,      Direction.Top],
                            });
                            _fields[r, q].ResourceType = GetNextResourceType(possibleResources);
                            if (_fields[r, q].ResourceType != ResourceType.None)
                                _fields[r, q].FieldNum = GetNextFieldNum(possibleNums);
                        }
                        else
                        {
                            _fields[r, q] = new Field();
                            _fields[r, q].ResourceType = ResourceType.Sea;
                        }

                        if (_fields[r, q].ResourceType == ResourceType.None)
                            RobberPosition = (r, q);
                    }
                }
            }
        }

        /// <summary>
        /// Checks if the given coordinates represent the inner part of the board.
        /// </summary>
        /// <returns>True if the given coordinates represent the inner part of the board, false otherwise.</returns>
        private bool IsInnerBoard(int r, int q)
        {
           return Math.Abs(q - Size / 2 + r - Size / 2) < 1 + (Size - 2) / 2 && q != 0 && q != Size - 1 && r != 0 && r != Size - 1;
        }

        private void InitializeNodes()
        {
            _nodes = new NodeMatrix(Size);
        }

        private void InitializeRoads()
        {
            _roads = new RoadMatrix(Size);
        }

        private void InitializePlayers()
        {
            _currentPlayer = 0;
            _turnIncrement = 1;
            _initialTurn = true;
            _rolled = true;
            _robberPositionChanged = true;
            _players = new Player[3];
            for (int i = 0; i < _players.Length; ++i)
            {
                _players[i] = new Player();
                _players[i].Id = (byte) i;
            }
        }

        /// <summary>
        /// Randomly selects a resource type from the given pool for initialization.
        /// It also takes one unit of the resource types out of the pool.
        /// </summary>
        /// <param name="possibleResources">The pool of resource types</param>
        /// <returns>The randomly selected resource type.</returns>
        private ResourceType GetNextResourceType(Dictionary<ResourceType, int> possibleResources)
        {
            List<ResourceType> possibilitiesList = possibleResources.Keys.ToList();
            ResourceType randResource = possibilitiesList[rand.Next(possibilitiesList.Count)];
            --possibleResources[randResource];

            if (possibleResources[randResource] == 0)
                possibleResources.Remove(randResource);

            return randResource;
        }

        /// <summary>
        /// Randomly selects a field number from the given pool for initialization.
        /// It also takes one unit of the field numbers out of the pool.
        /// </summary>
        /// <param name="possibleResources">The pool of field numbers.</param>
        /// <returns>The randomly selected field number.</returns>
        private int GetNextFieldNum(Dictionary<int, int> possibleResources)
        {
            List<int> possibilitiesList = possibleResources.Keys.ToList();
            int randNum = possibilitiesList[rand.Next(possibilitiesList.Count)];
            --possibleResources[randNum];

            if (possibleResources[randNum] == 0)
                possibleResources.Remove(randNum);

            return randNum;
        }

        /// <summary>
        /// Gets the coordinates of the nodes neighboring the node at the given coordinates.
        /// </summary>
        /// <param name="node">The coordinates of the given node.</param>
        /// <returns>The coordinates of the neighbors of the given node.</returns>
        /// <exception cref="InvalidOperationException">Gets thrown if an invalid direction coordinate is given in the NodeCoordinate.</exception>
        private List<NodeCoordinates> GetNeighborCoords(NodeCoordinates node)
        {
            List<NodeCoordinates> answer = new();
            switch (node.Dir)
            {
                case Direction.Top:
                    {
                        if (node.R > 0)
                            answer.Add(new NodeCoordinates(node.R - 1, node.Q, Direction.Bottom));
                        if (node.R > 0 && node.Q < 6)
                            answer.Add(new NodeCoordinates(node.R - 1, node.Q + 1, Direction.Bottom));
                        if (node.R > 1 && node.Q < 6)
                            answer.Add(new NodeCoordinates(node.R - 2, node.Q + 1, Direction.Bottom));
                        return answer;
                    }
                case Direction.Bottom:
                    {
                        if (node.R < 6)
                            answer.Add(new NodeCoordinates(node.R + 1, node.Q    , Direction.Top));
                        if (node.Q > 0 && node.R < 6)
                            answer.Add(new NodeCoordinates(node.R + 1, node.Q - 1, Direction.Top));
                        if (node.R < 5 && node.Q > 0)
                            answer.Add(new NodeCoordinates(node.R + 2, node.Q - 1, Direction.Top));
                        return answer;
                    }
                default:
                    throw new InvalidOperationException("Invalid Node direction!");
            }
        }

        /// <summary>
        /// Calculates the length of the road of the current player.
        /// Also assings max road to the current player if the road is atleast 5 long or longer than the previous longest road.
        /// </summary>
        private void MaxLengthRoadSearch()
        {
            int answer = 0;
            for (int i = 0; i < Size; i++)
            {
                for (int j = 0; j < Size; j++)
                {
                    for (int k = 0; k < 2; k++)
                    {
                        List<Road> visited = new();
                        NodeCoordinates from = new NodeCoordinates(i, j, (Direction)k);
                        answer = Math.Max(answer, RoadLength(from, visited));
                    }
                }
            }
            if (answer > 4 && answer > _maxRoadLength)
            {
                if (_maxLengthRoadOwnerId != -1)
                    _players[_maxLengthRoadOwnerId].Score -= 2;
                _players[CurrentPlayerId].Score += 2;
                _maxRoadLength = answer;
                _maxLengthRoadOwnerId = CurrentPlayerId;
            }
        }

        /// <summary>
        /// Calculates the longest road starting from the node with the given coordinate.
        /// </summary>
        /// <param name="from">The coordinates of the node from which the count starts.</param>
        /// <param name="visited">The list of roads already visited.</param>
        /// <param name="length">The length of the currently checked node.</param>
        /// <returns>The length of the longest road from the given node.</returns>
        private int RoadLength(NodeCoordinates from, List<Road> visited, int length = 0)
        {
            List<Road> visitedCopy = new List<Road>(visited);
            List<Road> roadsFrom = _roads.RoadsFromNode(from);
            int roadLength = length;
            foreach (Road r in roadsFrom)
            {
                if (r.PlayerId == CurrentPlayerId && !visitedCopy.Any(road => (road.Coordinates.From.Equals(r.Coordinates.From) && road.Coordinates.To.Equals(r.Coordinates.To)) 
                                                                           || (road.Coordinates.From.Equals(r.Coordinates.To) && road.Coordinates.To.Equals(r.Coordinates.From))))
                {
                    visitedCopy.Add(r);
                    roadLength = Math.Max(roadLength, RoadLength(r.Coordinates.To, visitedCopy, length + 1));
                }
            }
            return roadLength;
        }
        
        /// <summary>
        /// Assigns the largest army trait to the player who has enough Knight cards and has the most knight cards.
        /// </summary>
        private void MaxKnightCard()
        {
            if (Players[CurrentPlayerId].DevelopmentCards[CardType.Knight] > 2 && Players[CurrentPlayerId].DevelopmentCards[CardType.Knight] > _maxKnightCardCount)
            {
                if (_maxKnightCardOwnerId != -1)
                    Players[_maxKnightCardOwnerId].Score -= 2;
                Players[CurrentPlayerId].Score += 2;
                _maxKnightCardOwnerId = CurrentPlayerId;
                _maxKnightCardCount = Players[CurrentPlayerId].DevelopmentCards[CardType.Knight];
            }
        }

        /// <summary>
        /// Checks if an offer is valid.
        /// </summary>
        /// <param name="offers">The list of offers.</param>
        /// <returns>True if the offer is valid, false otherwise.</returns>
        private bool IsValidOffer(List<(ResourceType OfferedResource, int Quantity)> offers)
        {
            return offers.All(offer => offer.Quantity <= CurrentPlayer.Resources[offer.OfferedResource] && offer.Quantity % 3 == 0);
        }

        #endregion
    }
}
