﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CatanGame.Model
{
    /// <summary>
    /// The main class of the model layer.
    /// </summary>
    public class GameModel
    {
        #region fields
        private int _decies1 = 2;
        private int _decies2 = 1;
        #endregion

        #region Events

        /// <summary>
        /// Event: Is sent when a node is upgraded (a settlement or city is built).
        /// </summary>
        public event EventHandler<NodeCoordinates> NodeUpgraded;
        /// <summary>
        /// Event: Is sent when a road is placed.
        /// </summary>
        public event EventHandler<RoadPlacedEventArgs> RoadPlaced;
        /// <summary>
        /// Event: Is sent when the model sends which roads are available to build on.
        /// </summary>
        public event EventHandler<RoadsEnabledEventArgs> RoadsEnabled;
        /// <summary>
        /// Event: Is sent when the model sends which nodes are available to build on.
        /// </summary>
        public event EventHandler<List<NodeCoordinates>> NodesEnabled;
        /// <summary>
        /// Event: Is sent when the game is stepped.
        /// </summary>
        public event EventHandler GameStepped;
        /// <summary>
        /// Event: Is sent when the dice are rolled.
        /// </summary>
        public event EventHandler<RollEventArgs> GameRolled;
        /// <summary>
        /// Event: Is sent when the robber is moved.
        /// </summary>
        public event EventHandler<RobberMovedEventArgs> RobberMoved;
        /// <summary>
        /// Event: Is sent when 7 is rolled and the robber needs to move.
        /// </summary>
        public event EventHandler<RobberRolledEventArgs> RobberRolled;
        /// <summary>
        /// Event: Is sent when the resources change in the game.
        /// </summary>
        public event EventHandler ResourcesChanged;
        /// <summary>
        /// Event: Is sent when a card is bought.
        /// </summary>
        public event EventHandler<BuyCardEventArgs> CardBought;
        /// <summary>
        /// Event: Is sent when the game is over.
        /// </summary>
        public event EventHandler GameOver;
        #endregion

        #region Public properties
        /// <summary>
        /// Gets the result of the first die roll.
        /// </summary>
        public int Decies1 { get { return _decies1; } }
        /// <summary>
        /// Gets the result of the second die roll.
        /// </summary>
        public int Decies2 { get { return _decies2; } }
        /// <summary>
        /// Gets whether the game state is in the initial turn.
        /// </summary>
        public bool InitialTurn { get => Board.InitialTurn; }
        /// <summary>
        /// Gets whether the current player has rolled yet.
        /// </summary>
        public bool Rolled { get => Board.Rolled; }
        /// <summary>
        /// Gets or sets the board representing the game.
        /// </summary>
        public CatanBoard Board { get; private set; }
        /// <summary>
        /// Gets the id of the current player.
        /// </summary>
        public int CurrentPlayerId { get => Board.CurrentPlayerId; }
        #endregion

        #region Private methods
        /// <summary>
        /// Gets the coordinates of a node from a given node id.
        /// </summary>
        /// <param name="nodeId">The id of the node.</param>
        /// <returns>The coordinates of the node.</returns>
        private NodeCoordinates GetCoordinates(int nodeId)
        {
            int r = nodeId / (Board.Size * 2);
            int q = (nodeId / 2) % Board.Size;
            Direction dir = (Direction)((nodeId) % 2);
            return new NodeCoordinates(r, q, dir);
        }
        #endregion

        #region Public methods

        /// <summary>
        /// Starts a new game.
        /// </summary>
        public void NewGame()
        {
            Board = new CatanBoard();
        }

        /// <summary>
        /// Attempts to buy a card. Invokes CardBought event.
        /// </summary>
        public void BuyCard()
        {
            //int result = -3;
            CardType result;
            result = Board.BuyCard(Board.Players[CurrentPlayerId]);
            CardBought?.Invoke(this, new BuyCardEventArgs(result));

            IsGameOver();
        }

        /// <summary>
        /// Sends the available roads a given player can build through RoadsEvabled event.
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        public void SendEnabledRoads(int playerId)
        {
            RoadsEnabled?.Invoke(this, new RoadsEnabledEventArgs(Board.EnabledRoads(Board.Players[playerId])));
        }

        /// <summary>
        /// Sends the available nodes a given player can build through NodesEnabled event.
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        public void SendEnabledNodes(int playerId)
        {
            NodesEnabled?.Invoke(this, Board.EnableNodes(Board.Players[playerId]));
        }

        /// <summary>
        /// Checks whether the given player can build roads.
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        /// <returns>True if building roads is possible for the player, false otherwise.</returns>
        public bool CanBuildRoad(int playerId)
        {
            return Board.EnabledRoads(Board.Players[playerId]).Count > 0;
        }

        /// <summary>
        /// Checks whether the given player can build settlements..
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        /// <returns>True if building settlements is possible for the player, false otherwise.</returns>
        public bool CanBuildSettlement(int playerId)
        {
            return Board.EnableNodes(Board.Players[playerId]).Count > 0;
        }

        /// <summary>
        /// Checks whether the given player can buy a development card.
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        /// <returns>True if the player can buy a development card, false otherwise.</returns>
        public bool CanBuyDevelopmentCard(int playerId)
        {
            return !InitialTurn && Rolled && Board.RobberPlaced && Board.Players[playerId].CanBuyDevelopmentCard;
        }

        /// <summary>
        /// Checks whether a given player can trade.
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        /// <returns>True if the player can trade, false otherwise.</returns>
        public bool CanTrade(int playerId)
        {
            return !InitialTurn && Rolled && Board.RobberPlaced;
        }

        /// <summary>
        /// Checks whether a given player can step the game.
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        /// <returns>True if the player can step the game, false otherwise.</returns>
        public bool CanStep(int playerId)
        {
            return !InitialTurn && Rolled && Board.RobberPlaced;
        }

        /// <summary>
        /// Checks whether a given player can roll the dice.
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        /// <returns>True if the player can roll the dice, false otherwise.</returns>
        public bool CanRoll(int playerId)
        {
            return !InitialTurn && !Rolled;
        }

        /// <summary>
        /// Sends the roads the given player can build on that are adjacent with the given node.
        /// </summary>
        /// <param name="playerId">The id of the player.</param>
        /// <param name="pos">The coordinates of the node.</param>
        public void SendEnabledRoadsFromNode(int playerId, NodeCoordinates pos)
        {
            RoadsEnabled?.Invoke(this, new RoadsEnabledEventArgs(Board.EnabledRoadsFromNode(Board.Players[playerId], pos)));
        }

        /// <summary>
        /// Places a road.
        /// </summary>
        /// <param name="pos">The coordinates of the road.</param>
        public void PlaceRoad((int FirstNodeId, int SecondNodeId) pos)
        {
            int currPlayer = CurrentPlayerId;
            NodeCoordinates firstNodeCoords = GetCoordinates(pos.FirstNodeId);
            NodeCoordinates secondNodeCoords = GetCoordinates(pos.SecondNodeId);

            if (!Board.PlaceRoad(new RoadCoordinates(firstNodeCoords, secondNodeCoords), Board.Players[CurrentPlayerId]))
                return;

            RoadPlaced?.Invoke(this, new RoadPlacedEventArgs() {
                FirstNode = firstNodeCoords, 
                SecondNode = secondNodeCoords,
                OwnerId = currPlayer,
            });

            if (Board.InitialTurn)
            {
                GameStepped?.Invoke(this, EventArgs.Empty);
                SendEnabledNodes(CurrentPlayerId);
            }

            IsGameOver();
        }

        /// <summary>
        /// Checks whether the game is over. Invokes GameOver event if the game is over..
        /// </summary>
        /// <returns>True if the game is over, false otherwise.</returns>
        public bool IsGameOver()
        {
            if (Board.Players[CurrentPlayerId].Score > 4)
            {
                GameOver?.Invoke(this, new EventArgs());
                return true;
            }
            return false;
        }

        /// <summary>
        /// Attempts to upgrade a node(build a settlement if the node is empty, build city if not). 
        /// Invokes NodeUpgraded on success.
        /// </summary>
        /// <param name="node">The coordinates of the node.</param>
        /// <param name="playerId">The id of the player.</param>
        /// <returns></returns>
        public bool UpgradeNode(NodeCoordinates node, int playerId)
        {
            bool canPlace;
            if(Board.Nodes[node].Building.Type == HouseType.NoBuilding)
                canPlace = Board.PlaceSettlement(node, Board.Players[playerId]);
            else
                canPlace = Board.UpgradeToCity(node, Board.Players[playerId]);

            if (!canPlace)
                return false;

            NodeUpgraded?.Invoke(this, node);

            if (Board.InitialTurn)
                SendEnabledRoadsFromNode(playerId, node);

            IsGameOver();

            return true;
        }

        /// <summary>
        /// Moves the robber to the given coordinates.
        /// </summary>
        /// <param name="pos">The coordinates of the field to move the robber to.</param>
        public void MoveRobber((int R, int Q) pos)
        {
            (int R, int Q) oldRobberPos = Board.RobberPosition;
            if (Board.MoveRobber(pos))
                RobberMoved?.Invoke(this, new RobberMovedEventArgs(oldRobberPos, Board.RobberPosition));
        }

        /// <summary>
        /// Steps the game. Invokes GameStepped. 
        /// </summary>
        public void StepGame()
        {
            Board.StepGame();
            GameStepped?.Invoke(this, EventArgs.Empty);
        }

        /// <summary>
        /// Attempts to do a trade. Invokes ResourcesChanged on success.
        /// </summary>
        /// <param name="offers">The resources the player offers.</param>
        /// <param name="requests">The resources the player requests.</param>
        /// <returns>True if the trade was successful, false otherwise.</returns>
        public bool Trade(List<(ResourceType OfferedResource, int Quantity)> offers, List<(ResourceType RequestedResource, int Quantity)> requests)
        {
            bool result = Board.Trade(offers, requests);
            if (result)
                ResourcesChanged?.Invoke(this, new EventArgs());
            return result;
        }

        /// <summary>
        /// Checks whether a trade can be made.
        /// </summary>
        /// <param name="offers">The resources the player offers.</param>
        /// <param name="requests">The resources the player requests.</param>
        /// <returns>True if the player can make a trade, false otherwise.</returns>
        public bool CanTrade(List<(ResourceType OfferedResource, int Quantity)> offers, List<(ResourceType RequestedResource, int Quantity)> requests)
        {
            return Board.CanTrade(offers, requests);
        }

        /// <summary>
        /// Checks whether a player can request anymore resources.
        /// </summary>
        /// <param name="offers">The resources the player offers.</param>
        /// <param name="requests">The resources the player already requests.</param>
        /// <returns>True if the player can request more resources, false otherwise.</returns>
        public bool CanRequest(List<(ResourceType OfferedResource, int Quantity)> offers, List<(ResourceType RequestedResource, int Quantity)> requests)
        {
            return Board.CanRequest(offers, requests);
        }

        /// <summary>
        /// Rolls the dice. Invokes GameRolled.
        /// </summary>
        public void Roll()
        {
            if (!Board.Rolled)
            {
                (int FstDice, int SndDice) rollResult = Board.Roll();
                int rollSum = rollResult.FstDice + rollResult.SndDice;
                SendResources(rollSum);
                (_decies1, _decies2) = rollResult;
                GameRolled?.Invoke(this, new RollEventArgs(_decies1, _decies2));

                if (rollSum == 7)
                    SendRobberPositions();
            }
        }

        /// <summary>
        /// Sends the available positions for the robber, invoking RobberRolled.
        /// </summary>
        private void SendRobberPositions()
        {
            RobberRolled?.Invoke(this, new RobberRolledEventArgs(Board.PossibleRobberPositions()));
        }

        /// <summary>
        /// Sends the resources of the fields with the given number to the players who have settlements next to them.
        /// </summary>
        /// <param name="num">The field number that got rolled.</param>
        private void SendResources(int num)
        {
            if (num != 7)
                Board.SendResources(num);
        }

        #endregion
    }
}
