﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CatanGame.Model
{
    /// <summary>
    /// The matrix representing nodes in the board.
    /// </summary>
    public class NodeMatrix
    {
        private Node[,,] _nodes;

        /// <summary>
        /// Gets or sets the board size.
        /// </summary>
        public int BoardSize { get; private set; }

        public NodeMatrix(int boardSize)
        {
            BoardSize = boardSize;
            _nodes = new Node[BoardSize, BoardSize, 2];
            for (int q = 0; q < BoardSize; ++q)
            {
                for (int r = 0; r < BoardSize; ++r)
                {

                    if (IsInnerBoard(r, q))
                    {
                        _nodes[r, q, (int)Direction.Bottom] = new();
                        _nodes[r, q, (int)Direction.Bottom].Building = new(null);
                        _nodes[r, q, (int)Direction.Top] = new();
                        _nodes[r, q, (int)Direction.Top].Building = new(null);
                    }
                    else if (Math.Abs(q + r - BoardSize + 1) < 1 + BoardSize / 2)
                    {
                        if (r > BoardSize / 2)
                        {
                            _nodes[r, q, (int)Direction.Top] = new();
                            _nodes[r, q, (int)Direction.Top].Building = new(null);
                        }
                        else if (r < BoardSize / 2)
                        {
                            _nodes[r, q, (int)Direction.Bottom] = new();
                            _nodes[r, q, (int)Direction.Bottom].Building = new(null);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Checks whether the given coordinates are part of the inner board.
        /// </summary>
        /// <param name="r">The R axial coordinate.</param>
        /// <param name="q">The Q axial coordinate.</param>
        /// <returns>True if the coordinates are part of the inner board, false otherwise.</returns>
        private bool IsInnerBoard(int r, int q)
        {
            return Math.Abs(q - BoardSize / 2 + r - BoardSize / 2) < 1 + (BoardSize - 2) / 2 && q != 0 && q != BoardSize - 1 && r != 0 && r != BoardSize - 1;
        }

        /// <summary>
        /// Checks whether the node at the given coordinates exists.
        /// </summary>
        /// <param name="coords">The coordinates of the node to check</param>
        /// <returns>True if the node exists, false otherwise.</returns>
        public bool NodeExists(NodeCoordinates coords)
        {
            return coords.R >= 0 
                && coords.Q >= 0 
                && coords.R < BoardSize 
                && coords.Q < BoardSize 
                && this[coords] is not null;
        }

        
        /// <summary>
        /// Gets the Node at the given coordinates.
        /// </summary>
        /// <param name="coords">The coordinates of the node.</param>
        /// <returns>The node at the given coordinates.</returns>
        public Node this[NodeCoordinates coords]
        {
            get => GetNode(coords);
            set => SetNode(coords, value);
        }

        /// <summary>
        /// Gets the Node at the given coordinates.
        /// </summary>
        /// <param name="r">The R axial coordinate.</param>
        /// <param name="q">The Q axial coordinate.</param>
        /// <param name="dir">The direction coordinate(Top or Bottom).</param>
        /// <returns>The node ath the given coordinate.</returns>
        public Node this[int r, int q, Direction dir]
        {
            get => GetNode(new NodeCoordinates(r, q, dir));
            set => SetNode(new NodeCoordinates(r, q, dir), value);
        }

        /// <summary>
        /// Gets the Node at the given coordinates.
        /// </summary>
        /// <param name="coords">The coordinates of the node.</param>
        /// <returns>The node at the given coordinates.</returns>
        private Node GetNode(NodeCoordinates coords)
        {
            return GetNode(coords.R, coords.Q, coords.Dir);
        }

        /// <summary>
        /// Gets the Node at the given coordinates.
        /// </summary>
        /// <param name="r">The R axial coordinate.</param>
        /// <param name="q">The Q axial coordinate.</param>
        /// <param name="dir">The direction coordinate(Top or Bottom).</param>
        /// <returns>The node at the given coordinate.</returns>
        private Node GetNode(int r, int q, Direction dir)
        {
            return _nodes[r, q, (int) dir];
        }

        /// <summary>
        /// Sets the node at the given coordinate.
        /// </summary>
        /// <param name="coords">The coordinate of the node.</param>
        /// <param name="value">The node.</param>
        private void SetNode(NodeCoordinates coords, Node value)
        {
            _nodes[coords.R, coords.Q, (int)coords.Dir] = value;
        }
    }
}
