﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CatanGame.Model
{
    /// <summary>
    /// Represents a player in the model.
    /// </summary>
    public class Player
    {
        private Dictionary<ResourceType, int> _resources;
        private Dictionary<CardType, int> _developmentCards;

        public Player()
        {
            AvailableSettlementCnt = 5;
            AvailableCityCnt = 4;
            AvailableRoadCnt = 15;
            _resources = new()
            {
                // large values added for test purposes only
                { ResourceType.Brick, 0 },
                { ResourceType.Lumber, 0 },
                { ResourceType.Wool, 0 },
                { ResourceType.Grain, 0 },
                { ResourceType.Ore, 0 },
            };
            _developmentCards = new()
            {
                { CardType.VictoryPoint, 0 },
                { CardType.Knight, 0 },
            };
        }

        /// <summary>
        /// manages resources after building a settlement.
        /// </summary>
        /// <param name="initialTurn">whether or not we are in the initial turn.</param>
        public void AddedSettlement(bool initialTurn = false)
        {
            AvailableSettlementCnt--;
            if (!initialTurn)
            {
                Resources[ResourceType.Brick]--;
                Resources[ResourceType.Lumber]--;
                Resources[ResourceType.Wool]--;
                Resources[ResourceType.Grain]--;
            }
        }

        /// <summary>
        /// Handles resources after upgrading to city.
        /// </summary>
        public void UpgradedToCity()
        {
            Resources[ResourceType.Grain] -= 3;
            Resources[ResourceType.Ore] -= 3;
            AvailableCityCnt--;
            AvailableSettlementCnt++;
        }

        /// <summary>
        /// Handles resources after a road being built.
        /// </summary>
        /// <param name="initialTurn">Whether or not we are in the initial turn.</param>
        public void RoadBuilt(bool initialTurn = false)
        {
            AvailableRoadCnt--;
            if (!initialTurn)
            {
                Resources[ResourceType.Brick]--;
                Resources[ResourceType.Lumber]--;
            }
        }

        /// <summary>
        /// Handle resources after a development card was bought.
        /// </summary>
        public void DevelopmentCardBought()
        {
            Resources[ResourceType.Ore]--;
            Resources[ResourceType.Grain]--;
            Resources[ResourceType.Wool]--;
        }

        /// <summary>
        /// Gets or sets the id of the player.
        /// </summary>
        public int Id { get; set; }
        /// <summary>
        /// Gets or sets the score of the player.
        /// </summary>
        public int Score { get; set; }
        /// <summary>
        /// Gets or sets the developmentCards of the player.
        /// </summary>
        public Dictionary<CardType, int> DevelopmentCards { get => _developmentCards; set => _developmentCards = value; }
        /// <summary>
        /// Gets or sets the resources of the player.
        /// </summary>
        public Dictionary<ResourceType, int> Resources { get => _resources; set => _resources = value; }
        /// <summary>
        /// Gets whether the player can build a settlement.
        /// </summary>
        public bool CanBuildSettlement { get => AvailableSettlementCnt > 0 && _resources[ResourceType.Lumber] > 0 && _resources[ResourceType.Brick] > 0 && _resources[ResourceType.Grain] > 0 && _resources[ResourceType.Wool] > 0; }
        /// <summary>
        /// Gets whether the player can build a city.
        /// </summary>
        public bool CanBuildCity { get => AvailableCityCnt > 0 && _resources[ResourceType.Grain] > 2 && _resources[ResourceType.Ore] > 2; }
        /// <summary>
        /// Gets whether the player can build a road.
        /// </summary>
        public bool CanBuildRoad { get => AvailableRoadCnt > 0 && _resources[ResourceType.Lumber] > 0 && _resources[ResourceType.Brick] > 0; }
        /// <summary>
        /// Gets whether the player can buy a development card.
        /// </summary>
        public bool CanBuyDevelopmentCard { get => Resources[ResourceType.Ore] > 0 && Resources[ResourceType.Grain] > 0 && Resources[ResourceType.Wool] > 0; }
        /// <summary>
        /// Gets or sets the amount of settlements the player can still build.
        /// </summary>
        public int AvailableSettlementCnt { get; private set; }
        /// <summary>
        /// Gets or sets the amount of cities the player can still build.
        /// </summary>
        public int AvailableCityCnt { get; private set; }
        /// <summary>
        /// Gets or sets the amount of roads the player can still build.
        /// </summary>
        public int AvailableRoadCnt { get; private set; }
    }
}
