﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data;
using System.Linq;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Shapes;
using CatanGame.Model;
using CatanGame.View;
using CatanGame.ViewModel;

namespace CatanGame
{
    /// <summary>
    /// Interaction logic for App.xaml
    /// </summary>
    public partial class App : Application
    {
        private MainWindow _mainWindow;
        private CatanViewModel _viewModel;
        private GameModel _model;
        private NewGameWindow _newGameWindow;
        private GameWindow _gameWindow;
        private HelpWindow _helpWindow;
        private GameOverWindow _gameOverWindow;
        private ExchangeWindow _exchangeWindow;
        private BonusCardWindow _bonusCardWindow;
       
        App()
        {
            Startup += new StartupEventHandler(App_Startup);
        }

        #region Application event handlers

        /// <summary>
        /// Handles app startup event.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="e">The event arguments.</param>
        private void App_Startup(object sender, StartupEventArgs e)
        {
            _mainWindow = new MainWindow();
            _model = new GameModel();
            _newGameWindow = new NewGameWindow();
            _gameWindow = new GameWindow();
            _helpWindow = new HelpWindow();
            _exchangeWindow = new ExchangeWindow();
            _bonusCardWindow = new BonusCardWindow();
            _gameOverWindow = new GameOverWindow();

            _viewModel = new CatanViewModel(_model);
            _viewModel.UpgradeNode += ViewModel_UpgradeNode;
            _viewModel.PlaceRoad += ViewModel_PlaceRoad;
            _viewModel.MoveRobber += ViewModel_MoveRobber;
            _viewModel.CardBuy += ViewModel_CardBuy;
            _model.CardBought += new EventHandler<BuyCardEventArgs>(Model_CardBought);

            _viewModel.TradeData.TradeAccepted += ViewModel_TradeData_TradeAccepted;
            _viewModel.TradeData.Cancel += (sender, args) => _mainWindow.Navigate(_gameWindow);

            _gameWindow.boardCanvas.ViewModel = _viewModel;

            _mainWindow.DataContext = _viewModel;
            _viewModel.NewGameView += (o, args) => { _mainWindow.Navigate(_gameWindow); };
            _viewModel.NewGameNewNames += (o, args) => { _mainWindow.Navigate(_newGameWindow); };
            _viewModel.HelpView += (o, args) => { _mainWindow.Navigate(_helpWindow); };
            _viewModel.GameView += (o, args) => { _mainWindow.Navigate(_gameWindow); };
            _viewModel.GameOver += (o, args) => { _mainWindow.Navigate(_gameOverWindow); };
            _viewModel.ExchangeView += (o, args) => { _mainWindow.Navigate(_exchangeWindow); };
            _viewModel.BonusCardView += (o, args) => { _mainWindow.Navigate(_bonusCardWindow); };
            _mainWindow.Navigate(_newGameWindow);
            _mainWindow.Show();
        }

        #endregion

        /// <summary>
        /// Event handler. Upgrades the node at given coordinates.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="coords">The coordinates of the node to upgrade.</param>
        private void ViewModel_UpgradeNode(object sender, NodeCoordinates coords)
        {
            _model.UpgradeNode(coords, _model.CurrentPlayerId);
        }

        /// <summary>
        /// Event handler. Places a road at the given coordinates.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="pos">The coordinates of the road to place. </param>
        private void ViewModel_PlaceRoad(object sender, (int FirstNodeId, int SecondNodeId) pos)
        {
            _model.PlaceRoad(pos);
        }

        /// <summary>
        /// Event handler. Moves the robber to given coordinates.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="pos">The new coordinates of the robber.</param>
        private void ViewModel_MoveRobber(object sender, (int R, int Q) pos)
        {
            _model.MoveRobber(pos);
        }

        /// <summary>
        /// Event handler. Buys a bonus card.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="e">Empty event argument.</param>
        private void ViewModel_CardBuy(object sender, EventArgs e)
        {
            _model.BuyCard();
        }

        /// <summary>
        /// Event handler. Gives feedback to the user on which bonus card was bought.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="e">The result of the card buy.</param>
        private void Model_CardBought(object sender, BuyCardEventArgs e)
        {
            if (e.Result == CardType.Knight)
            {
                string message = "A húzott kártya Lovag";
                string title = "Bónusz kártya";
                MessageBox.Show(message, title);
            }
            else if (e.Result == CardType.VictoryPoint)
            {
                string message = "A húzott kártya +1 pont";
                string title = "Bónusz kártya";
                MessageBox.Show(message, title);
            }
            else if (e.Result == CardType.None)
            {
                string message = "Jelenleg nem tudsz kártyát húzni!";
                string title = "Bónusz kártya";
                MessageBox.Show(message, title);
            }
        }

        /// <summary>
        /// Event handler. Executes the trade and navigates back to the game window.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="e">The data related to the trade.</param>
        private void ViewModel_TradeData_TradeAccepted(object sender, TradeAcceptedEventArgs e)
        {
            _model.Trade(e.OfferedResources, e.RequestedResources);
            _mainWindow.Navigate(_gameWindow);
        }
    }
}
