﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;
using CatanGame.ViewModel;
using System.ComponentModel;

namespace CatanGame.View
{
    /// <summary>
    /// Canvas subclass used to draw a Catan Board
    /// </summary>
    public class CatanBoardCanvas : Canvas
    {
        #region Private fields
        private CatanViewModel _viewModel;
        #endregion

        #region Public constructors
        public CatanBoardCanvas() : base()
        { 
            this.Loaded += (obj, args) => {
                InitDraw();
                SizeChanged += (obj, args) => DrawMap();
            };
        }
        #endregion

        #region Private fields


        private LineWithOutline[,] _roads;
        private Ellipse[,,] _settlements;
        private CatanField[,] _fields;
        #endregion

        #region Public Properties
        /// <summary>
        /// Gets the offset with which the board moved.
        /// </summary>
        public int Offset { get => _viewModel.BoardSize / 2; }
        /// <summary>
        /// Gets or sets the diamater of the clickable ellipses on the board.
        /// </summary>
        public double EllipseDiameter { get; set; }
        /// <summary>
        /// Gets the size of the window.
        /// Note: The size of the window will always be considered the size on the smaller axis.
        /// </summary>
        public double WindowSize { get => Math.Min(ActualHeight, ActualWidth); }

        /// <summary>
        /// Gets the distance from the center of the hexagon to an edge of the hexagon 
        /// </summary>
        public double HexagonRadius { get => 1.15*WindowSize / (_viewModel.BoardSize * 2); }

        /// <summary>
        /// Gets or sets the ViewModel.
        /// </summary>
        public CatanViewModel ViewModel 
        { 
            get => _viewModel;
            set
            {
                _viewModel = value;
                _roads = new LineWithOutline[_viewModel.BoardSize * _viewModel.BoardSize * 2, _viewModel.BoardSize * _viewModel.BoardSize * 2];
                _settlements = new Ellipse[_viewModel.BoardSize, _viewModel.BoardSize, 2];
                _fields = new CatanField[_viewModel.BoardSize, _viewModel.BoardSize];
                InitDraw();
            }
        }

        #endregion

        #region Private methods
        /// <summary>
        /// Performs the initialization of the drawable objects on the board.
        /// </summary>
        private void InitDraw()
        {
            if (_viewModel is null || HexagonRadius == 0) return;

            Children.Clear();

            for (int q = 0; q < _viewModel.BoardSize; ++q)
            {
                for (int r = 0; r < _viewModel.BoardSize; ++r)
                {
                    if (_viewModel.Fields[r * _viewModel.BoardSize + q] is not null)
                    {
                        Point position = CalculatePositionOnCanvas(r, q);

                        CreateField(position, (r, q));
                        CreateRoadSpots(position, (r, q));
                    }
                }
            }
            CreateBuildingSpots();
        }

        /// <summary>
        /// Updates the board's view.
        /// </summary>
        private void DrawMap()
        {
            if (_viewModel is null || HexagonRadius == 0) return;

            for (int q = 0; q < _viewModel.BoardSize; ++q)
            {
                for (int r = 0; r < _viewModel.BoardSize; ++r)
                {
                    if (_viewModel.Fields[r * _viewModel.BoardSize + q] is not null)
                    {
                        Point position = CalculatePositionOnCanvas(r, q);

                        UpdateField(position, (r, q));
                        UpdateRoadSpots(position, (r, q));
                    }
                }
            }
            UpdateBuildingSpots();
        }

        /// <summary>
        /// Updates the position and size of a field on the board.
        /// </summary>
        /// <param name="position">The new position of the center of the field on the board.</param>
        /// <param name="coords">The coordinates of the field in the matrix.</param>
        private void UpdateField(Point position, (int R, int Q) coords)
        {
            _fields[coords.R, coords.Q].Update(position, HexagonRadius);
        }
        /// <summary>
        /// Updates the position of the roads around a given field.
        /// Note: it only updates three adjacent to the bottom,left of the field, for easy iteration purposes.
        /// </summary>
        /// <param name="pos">The new position of the center of the field on the board. </param>
        /// <param name="coords">The coordinates of the field in the matrix.</param>
        private void UpdateRoadSpots(Point pos, (int R, int Q) coords)
        {
            int maxId = _viewModel.BoardSize * _viewModel.BoardSize * 2;
            double marginOffset = EllipseDiameter / 2;
            int[] nodeIds = new int[]
            {
                GetNodeId(coords.R - 1, coords.Q + 1, 0),
                GetNodeId(coords.R + 1, coords.Q, 1),
                GetNodeId(coords.R, coords.Q, 0),
                GetNodeId(coords.R + 1, coords.Q - 1, 1),
            };
            for (int i = 0; i < 3; ++i)
            {
                Point firstNode = HexagonFactory.HexCorner(pos, HexagonRadius, i);
                Point secondNode = HexagonFactory.HexCorner(pos, HexagonRadius, i + 1);
                
                int firstNodeId = nodeIds[i];
                int secondNodeId = nodeIds[i + 1];

                if (firstNodeId >= 0 && secondNodeId >= 0 && firstNodeId < maxId && secondNodeId < maxId)
                {
                    _roads[firstNodeId, secondNodeId].X1 = firstNode.X;
                    _roads[firstNodeId, secondNodeId].Y1 = firstNode.Y;
                    _roads[firstNodeId, secondNodeId].X2 = secondNode.X;
                    _roads[firstNodeId, secondNodeId].Y2 = secondNode.Y;
                }
                    
            }
        }

        /// <summary>
        /// Updates the spots of the buildings
        /// </summary>
        private void UpdateBuildingSpots()
        { 
            double ellipseMarginOffset = EllipseDiameter / 2;
            int[] cornerIds = { 2, 5 };
            foreach (NodeViewModel node in _viewModel.Nodes)
            {
                if (node is null) continue;

                (int R, int Q, int D) coords = node.Position;
                Point position = CalculatePositionOnCanvas(coords.R, coords.Q);
                Point cornerPos = HexagonFactory.HexCorner(position, HexagonRadius, cornerIds[coords.D]);

                _settlements[coords.R, coords.Q, coords.D].Margin = new Thickness(cornerPos.X - ellipseMarginOffset, cornerPos.Y - ellipseMarginOffset, 0, 0);
            }
        }

        /// <summary>
        /// Creates a field on the board.
        /// </summary>
        /// <param name="position">The center of the field.</param>
        /// <param name="coords">The coordinates of the field in the matrix.</param>
        private void CreateField(Point position, (int R, int Q) coords)
        {
            CatanField field = new CatanField(position, coords, HexagonRadius);
            FieldViewModel fieldVM = _viewModel.Fields[coords.R * _viewModel.BoardSize + coords.Q];


            field.Field.DataContext = fieldVM;
            field.Field.Style = this.Resources["FieldStyle"] as Style;
            field.Robber.DataContext = fieldVM;
            field.Robber.Style = this.Resources["RobberStyle"] as Style;
            field.Robber.MouseDown += 
                (obj, args) => _viewModel.Fields[coords.R * _viewModel.BoardSize + coords.Q].MoveRobberCommand.Execute(null);

            SetZIndex(field.Robber, 50);

            if (fieldVM.FieldType != "None" && fieldVM.FieldType != "Sea")
            {
                field.Text.Content = fieldVM.FieldNumber;
                field.Text.FontWeight = FontWeights.DemiBold;
                field.Text.FontSize = 18;
            }
            Children.Add(field.Field);
            Children.Add(field.Text);
            Children.Add(field.Robber);
            _fields[coords.R, coords.Q] = field;

        }

        /// <summary>
        /// Creates the position of the roads around a given field.
        /// Note: it only creates three adjacent to the bottom,left of the field, for easy iteration purposes.
        /// </summary>
        /// <param name="pos">The position of the center of the field on the board. </param>
        /// <param name="coords">The coordinates of the field in the matrix.</param>
        private void CreateRoadSpots(Point pos, (int R, int Q) coords) 
        {
            int maxId = _viewModel.BoardSize * _viewModel.BoardSize * 2;
            double marginOffset = EllipseDiameter / 2;
            int[] nodeIds = new int[]
            {
                GetNodeId(coords.R - 1, coords.Q + 1, 0),
                GetNodeId(coords.R + 1, coords.Q, 1),
                GetNodeId(coords.R, coords.Q, 0),
                GetNodeId(coords.R + 1, coords.Q - 1, 1),
            };
            for (int i = 0; i < 3; ++i)
            {
                
                int firstNodeId = nodeIds[i];
                int secondNodeId = nodeIds[i + 1];
                if (firstNodeId < 0 || secondNodeId < 0 || firstNodeId >= maxId || secondNodeId >= maxId)
                    continue;

                Point firstNode = HexagonFactory.HexCorner(pos, HexagonRadius, i);
                Point secondNode = HexagonFactory.HexCorner(pos, HexagonRadius, i + 1);

                LineWithOutline roadLine = new LineWithOutline();
                roadLine.X1 = firstNode.X;
                roadLine.Y1 = firstNode.Y;
                roadLine.X2 = secondNode.X;
                roadLine.Y2 = secondNode.Y;
                roadLine.Line.Style = this.Resources["RoadFillStyle"] as Style;
                roadLine.Outline.Style = this.Resources["RoadOutlineStyle"] as Style;

                SetZIndex(roadLine.Line, 99);
                SetZIndex(roadLine.Outline, 98);

                roadLine.Line.DataContext = _viewModel.Roads[firstNodeId][secondNodeId];
                roadLine.Outline.DataContext = _viewModel.Roads[secondNodeId][firstNodeId];
                _roads[firstNodeId, secondNodeId] = roadLine;
                _roads[secondNodeId, firstNodeId] = roadLine;
                MouseButtonEventHandler buildRoad = (obj, param) => _viewModel.Roads[firstNodeId][secondNodeId].BuildRoadCommand.Execute(null);

                _roads[firstNodeId, secondNodeId].Line.MouseDown += buildRoad;
                _roads[firstNodeId, secondNodeId].Outline.MouseDown += buildRoad;

                Children.Add(roadLine.Line);
                Children.Add(roadLine.Outline);
            }
        }

        /// <summary>
        /// Creates the building spots on the board.
        /// </summary>
        private void CreateBuildingSpots()
        {
            double ellipseMarginOffset = EllipseDiameter / 2;
            int[] cornerIds = { 2, 5 };
            foreach (NodeViewModel node in _viewModel.Nodes)
            {
                if (node is null) continue;

                (int R, int Q, int D) coords =  node.Position;
                Point position = CalculatePositionOnCanvas(coords.R, coords.Q);
                Point cornerPos = HexagonFactory.HexCorner(position, HexagonRadius, cornerIds[coords.D]);
                Ellipse ellipse = new Ellipse();
                ellipse.Margin = new Thickness(cornerPos.X - ellipseMarginOffset, cornerPos.Y - ellipseMarginOffset, 0, 0);

                SetZIndex(ellipse, 100);
                int coordinateId = GetNodeId(coords.R, coords.Q, coords.D);
                ellipse.DataContext = _viewModel.Nodes[coordinateId];
                ellipse.Style = this.Resources["NodeStyle"] as Style;
                ellipse.MouseDown += (obj, args) => node.UpgradeNodeCommand.Execute(null);
                _settlements[coords.R, coords.Q, coords.D] = ellipse;
                Children.Add(ellipse);
            }           
        }

        /// <summary>
        /// Calculates the position of the center of the field at given coordinates on the canvas.
        /// </summary>
        /// <param name="r">The axial R coordinate.</param>
        /// <param name="q">The axial Q coordinate.</param>
        /// <returns>The position of the center of the field at given coordinates.</returns>
        private Point CalculatePositionOnCanvas(int r, int q)
        {
            int offsetR = r - Offset;
            int offsetQ = q - Offset;

            double xCoord = ActualWidth / 2 + HexagonRadius * (Math.Sqrt(3) * offsetQ + Math.Sqrt(3) * offsetR * 0.5);
            double yCoord = ActualHeight / 2 + HexagonRadius * 1.5 * offsetR;
            return  new Point(xCoord, yCoord);
        }

        
        /// <summary>
        /// Gets the id of a node based on it's axial coordinates.
        /// </summary>
        /// <param name="r">The axial R coordinate.</param>
        /// <param name="q">The axial Q coordinate.</param>
        /// <param name="dir">The axial direction:Up or Down.</param>
        /// <returns>The id of the node at given coordinates.</returns>
        private int GetNodeId(int r, int q, int dir)
        {
            return r * _viewModel.BoardSize * 2 + q * 2 + dir;
        }

        #endregion
    }
}
