﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CatanGame.Model;

namespace CatanGame.ViewModel
{
    public class CatanViewModel : ViewModelBase
    {
        #region Private fields
        private readonly GameModel _model;
        private PlayerViewModel _currentPlayer;
        #endregion

        #region Public properties
        public ObservableCollection<FieldViewModel> Fields { get; set; }
        public ObservableCollection<ObservableCollection<RoadViewModel>> Roads { get; set; }
        public ObservableCollection<NodeViewModel> Nodes { get; set; }
        public ObservableCollection<PlayerViewModel> Players { get; set; }
        public TradeViewModel TradeData { get; set; }
        public ResourceType[] Resources { get; set; }
        public PlayerViewModel CurrentPlayer { get => _currentPlayer; set => ChangeProperty(ref _currentPlayer, value); }

        /// <summary>
        /// Larges army owner for the view
        /// </summary>
        public String LargestArmyOwner { 
            get
            {
                if (_model.Board.MaxKnightCardOwnerId == -1)
                    return "";
                return Players[_model.Board.MaxKnightCardOwnerId].Name;
            } 
        }
        /// <summary>
        /// longest road owner for the view
        /// </summary>
        public String LongestRoadOwner
        {
            get
            {
                if (_model.Board.LongestRoadOwnerId == -1)
                    return "";
                return Players[_model.Board.LongestRoadOwnerId].Name;
            }
        }

        public bool CanBuildRoad { get => !_model.InitialTurn && _model.CanBuildRoad(CurrentPlayer.Id); }

        public bool CanBuildSettlement { get => !_model.InitialTurn && _model.CanBuildSettlement(CurrentPlayer.Id); }

        public bool CanBuyDevelopmentCard { get => _model.CanBuyDevelopmentCard(CurrentPlayer.Id); }

        public bool CanStep { get => _model.CanStep(CurrentPlayer.Id); }

        public bool CanRoll { get => _model.CanRoll(CurrentPlayer.Id); }

        public bool CanTrade { get => _model.CanTrade(CurrentPlayer.Id); }

        public int BoardSize { get => _model.Board.Size; }
        public int NodeCount { get => _model.Board.VertexCount; }
        public int CurrentPlayerId { get => _model.CurrentPlayerId; }
        public Int32 Dicies1 { get { return _model.Decies1; } }
        public Int32 Dicies2 { get { return _model.Decies2; } }
        #endregion

        #region Public constructors
        /// <summary>
        /// Constructor for catan view model
        /// </summary>
        /// <param name="model"> teh model name</param>
        public CatanViewModel(GameModel model)
        {
            //model csatlakozatatasa
            _model = model;
            _model.NodeUpgraded += Model_NodeChanged;
            _model.RoadPlaced += Model_RoadPlaced;
            _model.RoadsEnabled += Model_RoadsEnabled;
            _model.NodesEnabled += Model_NodesEnabled;
            _model.GameStepped += new EventHandler(Model_GameStepped);
            _model.GameRolled += Model_Rolled;
            _model.RobberMoved += Model_RobberMoved;
            _model.RobberRolled += Model_RobberRolled;
            _model.CardBought += Model_CardBought;
            _model.ResourcesChanged += (sender, args) => RefreshDatas();

            _model.GameOver += new EventHandler(OnGameOver);
            //Cards = new ObservableCollection<BonusCardsViewModel>();

            // initialize misc viewModels
            TradeData = new TradeViewModel(_model);

            // start a new game
            Init();
        }



        #endregion

        #region Private methods

        /// <summary>
        /// new game for view model
        /// </summary>
        private void NewGame()
        {
            _model.NewGame();
            
            InitPlayers();
            InitFields();
            InitNodes();
            InitRoads();
            ShowEnabledNodes(null);

            RefreshDatas();
            
        }
        /// <summary>
        /// Initializes the commands
        /// </summary>
        private void Init()
        {
            // initialize commands
            ShowEnabledRoadsCommand = new DelegateCommand(ShowEnabledRoads);
            ShowEnabledNodesCommand = new DelegateCommand(ShowEnabledNodes);
            StepGameCommand = new DelegateCommand(GameStep);
            BuyCardCommand = new DelegateCommand(BuyCard);
            RollCommand = new DelegateCommand(GameRoll);
            // Navigalashoz commandok
            NewGameViewCommand = new DelegateCommand(param => { OnNewGameView(); });
            NewGameNewNamesCommand = new DelegateCommand(param => { OnNewGameNewNames(); });
            HelpViewCommand = new DelegateCommand(param => { OnHelpView(); });
            BackToTheGameCommand = new DelegateCommand(param => { OnGameView(); });
            BonusCardViewCommand = new DelegateCommand(param => { OnBonusCardView(); });
            ExchangeViewCommand = new DelegateCommand(param => { OnExchangeView(); });

            NewGame();
        }
        /// <summary>
        /// Initializes the players
        /// </summary>
        private void InitPlayers()
        {

            if (Players is null)
            {
                Players = new();
                foreach (var player in _model.Board.Players)
                    Players.Add(new());
            }
            
            for (int i = 0; i < _model.Board.Players.Length; i++)
            {
                Players[i].Id = _model.Board.Players[i].Id;
                Players[i].Score = _model.Board.Players[i].Score;
                Players[i].ResourceCnt = new ObservableDictionary<string, int>();
                Players[i].Cards = new ObservableCollection<BonusCardsViewModel>();

                foreach (var entry in _model.Board.Players[i].Resources)
                    Players[i].ResourceCnt[entry.Key.ToString()] = entry.Value;
            }
            CurrentPlayer = Players[CurrentPlayerId];
        }
        /// <summary>
        /// Initializes the fields
        /// </summary>
        private void InitFields()
        {
            Fields = new();
            FieldViewModel field = null;
            for (int r = 0; r < _model.Board.Size; ++r)
            {
                for (int q = 0; q < _model.Board.Size; ++q)
                {
                    field = null;
                    if (_model.Board.Fields[r, q] is not null)
                    {
                        (int R, int Q) coordsCopy = (r, q);
                        field = new FieldViewModel()
                        {
                            R = r,
                            Q = q,
                            FieldNumber = _model.Board.Fields[r, q].FieldNum,
                            FieldType = _model.Board.Fields[r, q].ResourceType.ToString(),
                            HasRobber = false,
                            CanPlaceRobber = false,
                            MoveRobberCommand = new DelegateCommand(param => MoveRobber?.Invoke(this, coordsCopy))
                        };
                    }
                    Fields.Add(field);
                }
            }
            (int robberR, int robberQ) = _model.Board.RobberPosition;
            Fields[robberR * BoardSize + robberQ].HasRobber = true;
        }
        /// <summary>
        /// Initializes the nodes
        /// </summary>
        private void InitNodes()
        {
            Nodes = new();
            for (int r = 0; r < _model.Board.Size; ++r)
            {
                for (int q = 0; q < _model.Board.Size; ++q)
                {
                    for (int dir = 0; dir < 2; ++dir)
                    {
                        NodeViewModel node = null;
                        NodeCoordinates coords = new NodeCoordinates(r, q, (Direction)dir);
                        Node boardNode = _model.Board.Nodes[coords];
                        if (boardNode is not null)
                        {
                            node = new NodeViewModel()
                            {
                                HouseType = boardNode.Building.Type.ToString(),
                                Position = (r, q, dir),
                                UpgradeNodeCommand = new DelegateCommand((param) => UpgradeNode?.Invoke(this, coords)),
                                HouseOwner = "",
                                HouseOwnerId = -1
                            };

                        }
                        Nodes.Add(node);
                    }
                }
            }
        }
        /// <summary>
        /// Initializes the roads
        /// </summary>
        private void InitRoads()
        {
            int totalNodeCount = _model.Board.Size * _model.Board.Size * 2;
            Roads = new();
            // Note: iterating vertically to make symmetric matrix share references of Roads
            for (int i = 0; i < totalNodeCount; ++i)
                Roads.Add(new());

            for (int j = 0; j < totalNodeCount; ++j)
            {
                for (int i = 0; i < totalNodeCount; ++i)
                {
                    if (j > i)
                        Roads[i].Add(Roads[j][i]);
                    else
                    {
                        int iCop = i;
                        int jCop = j;
                        Roads[i].Add(new RoadViewModel()
                        {
                            IsBuildable = false,
                            Position = (iCop, jCop),
                            BuildRoadCommand = new DelegateCommand((param) => PlaceRoad?.Invoke(this, (iCop, jCop))),
                            OwnerId = -1,
                        });
                    }
                }
            }
        }
        /// <summary>
        /// refresh datas in the view
        /// </summary>
        private void RefreshDatas()
        {
            foreach (var p in Players)
                p.Score = _model.Board.Players[p.Id].Score;
            foreach (var entry in _model.Board.Players[CurrentPlayerId].Resources)
                CurrentPlayer.ResourceCnt[entry.Key.ToString()] = entry.Value;
            OnPropertyChanged(nameof(Dicies1));
            OnPropertyChanged(nameof(Dicies2));
            OnPropertyChanged(nameof(CanBuildRoad));
            OnPropertyChanged(nameof(CanBuildSettlement));
            OnPropertyChanged(nameof(CanBuyDevelopmentCard));
            OnPropertyChanged(nameof(CanRoll));
            OnPropertyChanged(nameof(CanStep));
            OnPropertyChanged(nameof(CanTrade));
            OnPropertyChanged(nameof(LargestArmyOwner));
            OnPropertyChanged(nameof(LongestRoadOwner));
        }
        /// <summary>
        /// show the place of enable roads
        /// </summary>
        /// <param name="param"></param>
        private void ShowEnabledRoads(object param)
        {
            _model.SendEnabledRoads(CurrentPlayerId);
        }
        /// <summary>
        /// show the place of enable nodes
        /// </summary>
        /// <param name="param"> the name of nodes</param>
        private void ShowEnabledNodes(object param)
        {
            _model.SendEnabledNodes(CurrentPlayerId);
        }
        /// <summary>
        /// Buy card in a model and  refresh the view
        /// </summary>
        /// <param name="param"></param>
        private void BuyCard(object param)
        {
            _model.BuyCard();
            RefreshDatas();
        }
        /// <summary>
        /// step the game in a model and  refresh the view
        /// </summary>
        /// <param name="param"></param>
        private void GameStep(object param)
        {
            _model.StepGame();
            RefreshDatas();
        }
        /// <summary>
        /// roll the game in a model
        /// </summary>
        /// <param name="param"></param>
        private void GameRoll(object param)
        {
            
            _model.Roll();
        }
        /// <summary>
        /// Calculates the node id for given player and returns them
        /// </summary>
        /// <param name="r"></param>
        /// <param name="q"></param>
        /// <param name="dir"></param>
        /// <returns></returns>

        private int GetNodeId(int r, int q, int dir)
        {
            return r * BoardSize * 2 + q * 2 + dir;
        }
        /// <summary>
        /// Calculates the node id for given player and returns them
        /// </summary>
        /// <param name="r"></param>
        /// <param name="q"></param>
        /// <param name="dir"></param>
        /// <returns></returns>
        private int GetNodeId(NodeCoordinates coords)
        {
            return GetNodeId(coords.R, coords.Q, (int) coords.Dir);
        }

        #endregion

        #region Public commands
        /// <summary>
        ///command for starting a new game
        /// </summary>
        public DelegateCommand NewGameCommand { get; private set; }
        /// <summary>
        /// command for exit a game
        /// </summary>
        public DelegateCommand ExitCommand { get; private set; }
        /// <summary>
        /// command for buying a bonus card
        /// </summary>
        public DelegateCommand BuyCardCommand { get; private set; }
        /// <summary>
        /// command for steping the game
        /// </summary>
        public DelegateCommand StepGameCommand { get; private set; }
        /// <summary>
        /// command for rolling the game
        /// </summary>
        public DelegateCommand RollCommand { get; private set; }
        /// <summary>
        /// command for show the enable roads in a view
        /// </summary>
        public DelegateCommand ShowEnabledRoadsCommand { get; private set; }
        /// <summary>
        /// command for show the enable nodes in a view
        /// </summary>
        public DelegateCommand ShowEnabledNodesCommand { get; private set; }

        //for navigation
        /// <summary>
        /// command for navigate to new game view
        /// </summary>
        public DelegateCommand NewGameViewCommand { get; private set; }
        /// <summary>
        /// command for navigate to new game new names view
        /// </summary>
        public DelegateCommand NewGameNewNamesCommand { get; private set; }
        /// <summary>
        /// command for navigate to help view
        /// </summary>
        public DelegateCommand HelpViewCommand { get; private set; }
        /// <summary>
        /// command for navigate to game board view
        /// </summary>
        public DelegateCommand BackToTheGameCommand { get; private set; }
        /// <summary>
        /// command for navigate to exchange view
        /// </summary>
        public DelegateCommand ExchangeViewCommand { get; private set; }
        /// <summary>
        /// command for navigate to bonus card view
        /// </summary>
        public DelegateCommand BonusCardViewCommand { get; private set; }
        #endregion

        #region Private methods
        /// <summary>
        /// delet the enable roads and nodes in view
        /// </summary>
        private void ClearOptions()
        {
            ClearBuildableRoads();
            ClearBuildableNodes();
        }
        /// <summary>
        /// delet the enable roads in view
        /// </summary>
        private void ClearBuildableRoads()
        {
            foreach (var seconds in Roads)
                foreach (var road in seconds)
                    if (road is not null)
                        road.IsBuildable = false;
        }
        /// <summary>
        /// delet the enable nodes in view
        /// </summary>
        private void ClearBuildableNodes()
        {
            foreach (NodeViewModel node in Nodes)
                if (node is not null)
                    node.IsBuildable = false;
        }

        #endregion

        #region Model event handlers
        /// <summary>
        /// event handler for card bought
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="e">buy card event args</param>
        private void Model_CardBought(object obj, BuyCardEventArgs e)
        {
            if (e.Result == CardType.Knight)
            {
                CurrentPlayer.Cards.Add(new BonusCardsViewModel() {CardType="Knight" });
            }else if(e.Result == CardType.VictoryPoint)
            {
                CurrentPlayer.Cards.Add(new BonusCardsViewModel(){ CardType = "PlusPoint" });
            }
            
        }

        /// <summary>
        /// event handler for nodw changed
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="coords">the node cords</param>
        private void Model_NodeChanged(object obj, NodeCoordinates coords)
        {
            int nodeId = coords.R * BoardSize * 2 + coords.Q * 2 + (int)coords.Dir;
            Node node = _model.Board.Nodes[coords];

            Nodes[nodeId].HouseType = node.Building.Type.ToString();
            Nodes[nodeId].HouseOwnerId = node.Building.Owner.Id;

            RefreshDatas();

            ClearOptions();
        }
        /// <summary>
        /// event handler for  road placed
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="eventArgs">road place event args</param>
        private void Model_RoadPlaced(object obj, RoadPlacedEventArgs eventArgs)
        {
            Roads[GetNodeId(eventArgs.FirstNode)][GetNodeId(eventArgs.SecondNode)].OwnerId = eventArgs.OwnerId;

            RefreshDatas();

            ClearOptions();
        }
        /// <summary>
        /// event handler for robber moved
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="eventArgs">event args</param>
        private void Model_RobberMoved(object obj, RobberMovedEventArgs eventArgs)
        {
            Fields[eventArgs.From.R * BoardSize + eventArgs.From.Q].HasRobber = false;
            Fields[eventArgs.To.R * BoardSize + eventArgs.To.Q].HasRobber = true;
            foreach (FieldViewModel field in Fields)
                if (field is not null)
                    field.CanPlaceRobber = false;
            RefreshDatas();
        }
        /// <summary>
        /// event handler for game stepped
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="e">event args</param>
        private void Model_GameStepped(object obj, EventArgs e)
        {
            CurrentPlayer = Players[_model.CurrentPlayerId];
            ClearOptions();
        }
        /// <summary>
        /// event handler for the game rolled
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="e">roll event args</param>
        private void Model_Rolled(object obj, RollEventArgs e)
        {
            //GameRolled?.Invoke(this, new RollEventArgs(e.Decies1,e.Decies2));
            RefreshDatas();
        }
        /// <summary>
        /// event handler for rolled the robber
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="eventArgs">event args</param>
        private void Model_RobberRolled(object obj, RobberRolledEventArgs eventArgs)
        {
            foreach ((int R, int Q) coords in eventArgs.PossibleRobberPositions)
                Fields[coords.R * BoardSize + coords.Q].CanPlaceRobber = true;
        }
        /// <summary>
        /// event handler for new game
        /// </summary>
        private void OnNewGameView()
        {
            if (Players[0].Name != null && Players[1].Name != null && Players[2].Name != null)
            {
                Random rand = new();

                List<string> names = new List<string>() { Players[0].Name, Players[1].Name, Players[2].Name };

                for (int i = 0; i < 3; i++)
                {
                    int ind = rand.Next(0, names.Count);
                    Players[i].Name = names[ind];
                    names.RemoveAt(ind);
                }

                NewGameView?.Invoke(this, EventArgs.Empty);
            }
            NewGame();
        }
        /// <summary>
        /// event handler for new names
        /// </summary>
        private void OnNewGameNewNames()
        {
            NewGameNewNames?.Invoke(this, EventArgs.Empty);
        }
        /// <summary>
        /// event handler for navigate to help view
        /// </summary>
        private void OnHelpView()
        {
            HelpView?.Invoke(this, EventArgs.Empty);
        }
        /// <summary>
        /// event handler for navigate to bonus card view
        /// </summary>
        private void OnBonusCardView()
        {
            BonusCardView?.Invoke(this, EventArgs.Empty);
        }
        /// <summary>
        /// event handler for navigate to exchange view
        /// </summary>
        private void OnExchangeView()
        {
            TradeData.Update(CurrentPlayer);
            ExchangeView?.Invoke(this, EventArgs.Empty);
        }
        /// <summary>
        /// event handler for navigate to game over view
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="e">event args</param>
        private void OnGameOver(object obj, EventArgs e)
        {
            GameOver?.Invoke(this, EventArgs.Empty);
        }
        /// <summary>
        /// event handler for navigate to game board view
        /// </summary>
        private void OnGameView()
        {
            GameView?.Invoke(this, EventArgs.Empty);
        }
        /// <summary>
        /// event handler for show enable roads
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="eventArgs">event args</param>
        private void Model_RoadsEnabled(object obj, RoadsEnabledEventArgs eventArgs)
        {
            ClearOptions();

            // setting IsBuildable to true at the reverse coordinates is not needed, because 
            // the normal coordinates and the reverse coordinates share the same reference.
            eventArgs.EnabledRoads.ForEach((road) => {
                RoadViewModel roadVM = Roads[GetNodeId(road.From)][GetNodeId(road.To)];
                if (roadVM is not null)
                   roadVM.IsBuildable = true;
            });
        }
        /// <summary>
        /// event handler for show enable nodes
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="enabledNodes">the list of enable nodes</param>
        private void Model_NodesEnabled(object obj, List<NodeCoordinates> enabledNodes)
        {
            ClearOptions();

            enabledNodes.ForEach((node) => Nodes[GetNodeId(node)].IsBuildable = true);
        }


        #endregion

        #region Public events

        public event EventHandler<EventArgs> Exit;
        public event EventHandler<EventArgs> CardBuy;
        public event EventHandler<(int FirstNodeId, int SecondNodeId)> PlaceRoad;
        public event EventHandler<NodeCoordinates> UpgradeNode;
        public event EventHandler<EventArgs> UpgradeToCity;
        public event EventHandler<EventArgs> StepGame;
        public event EventHandler<EventArgs> Roll;
        public event EventHandler<RollEventArgs> GameRolled;
        public event EventHandler<(int R, int Q)> MoveRobber;
 
        // For navigation
        public event EventHandler<EventArgs> GameOver;
        public event EventHandler ExchangeView;
        public event EventHandler BonusCardView;
        public event EventHandler NewGameView;
        public event EventHandler NewGameNewNames;
        public event EventHandler HelpView;
        public event EventHandler GameView;

        #endregion
    }
}
