﻿using System;
using System.Windows.Input;

namespace CatanGame.ViewModel
{
    // Code taken from class resources
    public class DelegateCommand : ICommand
    {
        private readonly Action<Object> _execute; // the term lambda performing the activity
        private readonly Func<Object, Boolean> _canExecute; // the term lambda against the condition of activity

        /// <summary>
        /// Create a command.
        /// </summary>
        /// <param name="execute">Activity to be performed.</param>
        public DelegateCommand(Action<Object> execute) : this(null, execute) { }

        /// <summary>
        /// Create a command.
        /// </summary>
        /// <param name="canExecute">Condition of enforceability.</param>
        /// <param name="execute">Activity to be performed.</param>
        public DelegateCommand(Func<Object, Boolean> canExecute, Action<Object> execute)
        {
            if (execute == null)
            {
                throw new ArgumentNullException("execute");
            }

            _execute = execute;
            _canExecute = canExecute;
        }

        /// <summary>
        /// Event of change in enforceability
        /// </summary>
        public event EventHandler CanExecuteChanged;

        /// <summary>
        /// Verification of feasibility
        /// </summary>
        /// <param name="parameter">Activity parameter</param>
        /// <returns>True if the activity is executable.</returns>
        public Boolean CanExecute(Object parameter)
        {
            return _canExecute == null ? true : _canExecute(parameter);
        }

        /// <summary>
        ///Execution of an activity.
        /// </summary>
        /// <param name="parameter">Activity parameter.</param>
        public void Execute(Object parameter)
        {
            if (!CanExecute(parameter))
            {
                throw new InvalidOperationException("Command execution is disabled.");
            }
            _execute(parameter);
        }

        /// <summary>
        /// Triggering a change in executability event
        /// </summary>
        public void RaiseCanExecuteChanged()
        {
            if (CanExecuteChanged != null)
                CanExecuteChanged(this, EventArgs.Empty);
        }
    }
}
