﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Collections.Specialized;
using System.Collections;

namespace CatanGame.ViewModel
{
    /// <summary>
    /// An observable dictionary.
    /// </summary>
    /// <typeparam name="Key">The type of the keys in the dictionary.</typeparam>
    /// <typeparam name="Value">The type of the values in the dictionary.</typeparam>
    public class ObservableDictionary<Key, Value> : ViewModelBase, IEnumerable, INotifyCollectionChanged, IEquatable<ObservableDictionary<Key, Value>> where Value : IEquatable<Value>
    {
        private Dictionary<Key, Value> _dict;

        /// <summary>
        /// Event: notifies the view when the collection changes.
        /// </summary>
        public event NotifyCollectionChangedEventHandler CollectionChanged;

        /// <summary>
        /// gets the underlying dictionary.
        /// </summary>
        public Dictionary<Key, Value> Dict { get => _dict; }

        public ObservableDictionary()
        {
            _dict = new Dictionary<Key, Value>();
        }

        /// <summary>
        /// Gets or inserts a value at given key. Invokes CollectionChanged when changes happen.
        /// </summary>
        /// <param name="k">The key.</param>
        /// <returns>The value at given key.</returns>
        public Value this[Key k]
        {
            get => _dict[k];

            set
            {
                if (!_dict.ContainsKey(k))
                {
                    _dict[k] = value;
                    OnPropertyChanged("Key");
                    OnPropertyChanged("Value");
                    CollectionChanged?.Invoke(this, new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
                    return;
                }
                
                Value val = _dict[k];
                if (val is not null && val.Equals(value))
                    return;
                _dict[k] = value;
                OnPropertyChanged("Value");
                CollectionChanged?.Invoke(this, new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
            }
        }

        /// <summary>
        /// Checks whether this dictionary is equal to another dictionary.
        /// </summary>
        /// <param name="other">The other dictionary.</param>
        /// <returns>True if the equality holds, false otherwise.</returns>
        public bool Equals(ObservableDictionary<Key, Value> other)
        {
            return _dict.Equals(other);
        }

        /// <summary>
        /// Gets the enumerator of the dictionary.
        /// </summary>
        /// <returns>The enumerator of the dictionary.</returns>
        public IEnumerator GetEnumerator()
        {
            return _dict.GetEnumerator();
        }
    }
}
