﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CatanGame.Model;

namespace CatanGame.ViewModel
{
    /// <summary>
    /// ViewModel representing a trade.
    /// </summary>
    public class TradeViewModel : ViewModelBase
    {
        #region Private fields
        private GameModel _model;
        private ObservableDictionary<string, TradeDataViewModel> _tradeData;
        #endregion

        #region Public properties
        /// <summary>
        /// Gets or sets the Data related to trading.
        /// </summary>
        public ObservableDictionary<string, TradeDataViewModel> TradeData { get => _tradeData; set => ChangeProperty(ref _tradeData, value); }

        /// <summary>
        /// Gets whether a trade is possible.
        /// </summary>
        public bool CanTrade { get => _model.CanTrade(OfferedResources, RequestedResources); }

        /// <summary>
        /// Gets whether the player can request anymore resources.
        /// </summary>
        public bool CanRequest { get => _model.CanRequest(OfferedResources, RequestedResources); }
        #endregion

        #region Commands
        /// <summary>
        /// Command for accepting a trade.
        /// </summary>
        public DelegateCommand AcceptTradeCommand { get; private set; }
        /// <summary>
        /// Command for clearing the trade.
        /// </summary>
        public DelegateCommand ClearTradeCommand { get; private set; }

        /// <summary>
        /// Command for canceling a trade.
        /// </summary>
        public DelegateCommand CancelCommand { get; private set; }
        #endregion

        #region Public events
        /// <summary>
        /// Event: Gets invoked when a trade is accepted.
        /// </summary>
        public event EventHandler<TradeAcceptedEventArgs> TradeAccepted;
        /// <summary>
        /// Event: Gets invoked when a trade is canceled.
        /// </summary>
        public event EventHandler Cancel;
        #endregion

        #region Private properties
        /// <summary>
        /// Gets the resources offered by the player.
        /// </summary>
        private List<(ResourceType, int)> OfferedResources 
        {
            get => _tradeData.Dict.Select(data => ((ResourceType)Enum.Parse(typeof(ResourceType), data.Key), data.Value.OfferedCnt)).ToList();
        }
        /// <summary>
        /// Gets the resources requested by the player.
        /// </summary>
        private List<(ResourceType, int)> RequestedResources
        {
            get => _tradeData.Dict.Select(data => ((ResourceType)Enum.Parse(typeof(ResourceType), data.Key), data.Value.RequestedCnt)).ToList();
        }
        #endregion

        #region Public constructors
        public TradeViewModel(GameModel model)
        {
            TradeData = new();
            _model = model;
            CancelCommand = new DelegateCommand(obj => Cancel.Invoke(this, EventArgs.Empty));
            AcceptTradeCommand = new DelegateCommand(obj => TradeAccepted?.Invoke(this, new TradeAcceptedEventArgs(OfferedResources, RequestedResources)));
            ClearTradeCommand = new DelegateCommand(obj =>
            {
                foreach(var trade in _tradeData.Dict)
                {
                    trade.Value.OfferedCnt = 0;
                    trade.Value.RequestedCnt = 0;
                }    
            });
        }
        #endregion

        #region Public methods
        /// <summary>
        /// Updates this viewmodel with the data from the given player.
        /// </summary>
        /// <param name="player">The player who wants to make a trade.</param>
        public void Update(PlayerViewModel player)
        {
            TradeData.Dict.Clear();
            foreach (var pair in player.ResourceCnt.Dict)
            {
                TradeDataViewModel data = new TradeDataViewModel(pair.Value, 0, 0);
                data.TradeDataChanged += (sender, e) => OnPropertyChanged(nameof(CanTrade));
                data.TradeDataChanged += (sender, e) => OnPropertyChanged(nameof(CanRequest));
                TradeData[pair.Key] = data;
            }
        }
        #endregion
    }
}
